#include <list>
#include <map>
using namespace std;

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "tinyxml.h"

#include "outstreams.hpp"
#include "config.hpp"
#include "coordinates.hpp"
#include "SC2Map.hpp"
#include "sc2mapTypes.hpp"



//float SC2Map::defaultMineralAmount = 1500.0f;
//float SC2Map::defaultGeyserAmount  = 2500.0f;



struct FootToApply
{
  point  loc;
  string type;
  string name;
};


static list<point>       pathingFills;
static list<FootToApply> footsToApply;



void SC2Map::processPlacedObject( ObjectMode objMode, TiXmlElement* object )
{
  char* objPoint;
  char* objUnit;
  char* objDoodad;
  char* attType;
  char* attPosition;
  char* attUnitType;
  char* attResources;

  switch( objMode )
  {
    case OBJMODE_BETAPH1_v26:
      objPoint     = "Point";
      objUnit      = "Unit";
      objDoodad    = "Doodad";
      attType      = "type";
      attPosition  = "position";
      attUnitType  = "unitType";
      attResources = "resources";
    break;


    case OBJMODE_BETAPH2_v26:
      objPoint     = "ObjectPoint";
      objUnit      = "ObjectUnit";
      objDoodad    = "ObjectDoodad";
      attType      = "Type";
      attPosition  = "Position";
      attUnitType  = "UnitType";
      attResources = "Resources";
    break;


    default:
      printError( "Unknown object processing mode?\n" );
      exit( -1 );
  }


  // everything has a position
  const char* strPosition = object->Attribute( attPosition );
  float mx, my, mz;
  sscanf( strPosition, "%f,%f,%f", &mx, &my, &mz );



  // Points might be a start location
  if( strcmp( object->Value(), objPoint ) == 0 )
  {
    const char* strType = object->Attribute( attType );


    if( strcmp( strType, "StartLoc" ) == 0 )
    {
      StartLoc* sl = new StartLoc;
      sl->loc.mSet( mx, my );
      startLocs.push_back( sl );
      return;
    }


    if( strcmp( strType, "BlockPathing" ) == 0 )
    {
      point c;
      c.mSet( mx, my );
      pathingFills.push_back( c );
      return;
    }
  }



  // doodads might have a footprint, don't worry about scaled
  // doodads--their footprint is the same
  if( strcmp( object->Value(), objDoodad ) == 0 )
  {
    const char* strType = object->Attribute( attType );

    string type( "doodad" );
    string name( strType );

    FootToApply fta;
    fta.loc.mSet( mx, my );
    fta.type.assign( type );
    fta.name.assign( name );
    footsToApply.push_back( fta );
  }



  // units come in three flavors:
  //   resource - can be mined, has a footprint
  //   destruct - can be destroyed, has a footprint
  //   unit     - may have a footprint

  if( strcmp( object->Value(), objUnit ) == 0 )
  {
    // every unit has a position (TRUE?)
    const char* strPosition = object->Attribute( attPosition );
    float mx, my, mz;
    sscanf( strPosition, "%f,%f,%f", &mx, &my, &mz );

    // every unit has a unitType (TRUE?)
    const char* strType = object->Attribute( attUnitType );

    // might be used by resources below
    const char* strResources = object->Attribute( attResources );



    // an object may have more than one type of footprint,
    // all of which will be applied
    FootToApply fta;

    string name( strType );

    string type( "destruct" );
    fta.loc.mSet( mx, my );
    fta.type.assign( type );
    fta.name.assign( name );
    footsToApply.push_back( fta );

    type.assign( "unit" );
    fta.loc.mSet( mx, my );
    fta.type.assign( type );
    fta.name.assign( name );
    footsToApply.push_back( fta );

    type.assign( "resource" );
    fta.loc.mSet( mx, my );
    fta.type.assign( type );
    fta.name.assign( name );
    footsToApply.push_back( fta );

    type.assign( "nobuildmain" );
    fta.loc.mSet( mx, my );
    fta.type.assign( type );
    fta.name.assign( name );
    footsToApply.push_back( fta );



    if( strcmp( strType, "Observatory"  ) == 0 ||
        strcmp( strType, "XelNagaTower" ) == 0 )
    {
      Watchtower* wt = new Watchtower;
      wt->loc.mSet( mx, my );
      wt->range = 22.0f;
      watchtowers.push_back( wt );
      return;
    }


    if( strcmp( strType, "MineralField" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = MINERALS;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = getfConstant( "defaultMineralAmount" );
      }
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "RichMineralField"      ) == 0 ||
        strcmp( strType, "HighYieldMineralField" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = MINERALS_HY;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = getfConstant( "defaultMineralAmount" );
      }
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "VespeneGeyser"       ) == 0 ||
        strcmp( strType, "SpacePlatformGeyser" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = VESPENEGAS;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = getfConstant( "defaultGeyserAmount" );
      }
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "RichVespeneGeyser" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = VESPENEGAS_HY;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = getfConstant( "defaultGeyserAmount" );
      }
      resources.push_back( r );
      return;
    }
  }
}




// From editor testing:
// dynamic pathing fills flow across ramps!
// they do NOT flow across painted green pathing
// they DO flow across (ignore) doodads
// THEREFORE, the dynamic fills should be executed before
// any placed objects modify the pathing map made from
// the terrain only
//
// SPECIAL NOTE: as if this wasn't already complicated--this
// tool already applies the painted pathing layer BEFORE doing
// placed objects so we can tell when resources are in unpathable
// areas, which means we can't do dynamic fills BEFORE painted
// pathing is applied.  The way to really handle this is to keep
// painted pathing as a separate layer up to this point and apply
// after, then to do resources you check if they are in the
// terrain pathing OR in the painted pathing layer
void SC2Map::applyFillsAndFootprints()
{
  for( list<point>::iterator itr = pathingFills.begin();
       itr != pathingFills.end();
       ++itr )
  {
    applyFill( &(*itr) );
  }
  pathingFills.clear();


  for( list<FootToApply>::iterator itr = footsToApply.begin();
       itr != footsToApply.end();
       ++itr )
  {
    FootToApply* fta = &(*itr);
    applyFootprint( &(fta->loc), &(fta->type), &(fta->name) );
  }
  footsToApply.clear();
}



void SC2Map::applyFill( point* src )
{
  if( !isPlayableCell( src ) )
  {
    return;
  }


  map<int, point> fillSet;
  map<int, point> workSet;

  workSet[makePointKey( src )] = *src;


  while( !workSet.empty() )
  {
    map<int, point>::iterator itr = workSet.begin();
    point c = itr->second;
    workSet.erase( makePointKey( &c ) );

    fillSet[makePointKey( &c )] = c;

    //propagateFill(  1,  1, &c, &fillSet, &workSet );
    //propagateFill(  1,  0, &c, &fillSet, &workSet );
    //propagateFill(  1, -1, &c, &fillSet, &workSet );
    //propagateFill(  0, -1, &c, &fillSet, &workSet );
    //propagateFill( -1, -1, &c, &fillSet, &workSet );
    //propagateFill( -1,  0, &c, &fillSet, &workSet );
    //propagateFill( -1,  1, &c, &fillSet, &workSet );
    //propagateFill(  0,  1, &c, &fillSet, &workSet );
  }


  for( map<int, point>::iterator itr = fillSet.begin();
       itr != fillSet.end();
       ++itr )
  {
    point c = itr->second;

    for( int t = 0; t < NUM_PATH_TYPES; ++t )
    {
      setPathing( &c, (PathType)t, false );
    }
  }
}


int SC2Map::makePointKey( point* c )
{
  return c->pcx + c->pcy*cxDimPlayable;
}


void SC2Map::propagateFill( int dx, int dy, point* c, map<int, point>* fillSet, map<int, point>* workSet )
{
  point cn;
  cn.pcSet( c->pcx + dx, c->pcy + dy );

  if( !isPlayableCell( &cn ) )
  {
    return;
  }

  if( !getPathing( &cn, PATH_GROUND_NOROCKS ) )
  {
    return;
  }

  int k = makePointKey( &cn );

  map<int, point>::iterator itr = fillSet->find( k );

  if( itr == fillSet->end() )
  {
    (*workSet)[k] = cn;
  }
}





void SC2Map::applyFootprint( point* c, string* type, string* name )
{
  Footprint* foot = NULL;;

  map<string, Footprint*>* innerMap;


  // first check local user config
  innerMap = configUserLocal.type2name2foot[*type];
  if( innerMap != NULL )
  {
    foot = (*innerMap)[*name];
  }


  // if nothing there, check global
  if( foot == NULL )
  {
    innerMap = configUserGlobal.type2name2foot[*type];
    if( innerMap != NULL )
    {
      foot = (*innerMap)[*name];
    }
  }


  // note: there are NO FOOTPRINTS in the global internal
  // config, so if we don't have one by now, just quit
  if( foot == NULL )
  {
    return;
  }

  for( list<int>::iterator itr = foot->relativeCoordinates.begin();
       itr != foot->relativeCoordinates.end();
       ++itr )
  {
    int dx = *itr;

    // there should be an even number of coordinates!
    ++itr;
    if( itr == foot->relativeCoordinates.end() )
    {
      printError( "Odd number of coordinates for footprint %s %s.\n",
                  type->data(),
                  name->data() );
      exit( -1 );
    }

    int dy = *itr;

    point dc;
    dc.pcSet( c->pcx + dx, c->pcy + dy );

    if( !isPlayableCell( &dc ) )
    {
      continue;
    }

    if( *type == "doodad" || *type == "unit" ) {
      // remove all pathing
      setPathing( &dc, PATH_GROUND_NOROCKS,               false );
      setPathing( &dc, PATH_GROUND_WITHROCKS,             false );
      setPathing( &dc, PATH_CWALK_NOROCKS,                false );
      setPathing( &dc, PATH_CWALK_WITHROCKS,              false );
      setPathing( &dc, PATH_GROUND_WITHROCKS_NORESOURCES, false );
      setPathing( &dc, PATH_BUILDABLE,                    false );
      setPathing( &dc, PATH_BUILDABLE_MAIN,               false );

    } else if( *type == "destruct" ) {
      // remove from WITHROCKS pathing types
      setPathing( &dc, PATH_GROUND_WITHROCKS,             false );
      setPathing( &dc, PATH_CWALK_WITHROCKS,              false );
      setPathing( &dc, PATH_GROUND_WITHROCKS_NORESOURCES, false );

    } else if( *type == "resource" ) {
      // remove all pathing except WITHOUTRESOURCES
      setPathing( &dc, PATH_GROUND_NOROCKS,               false );
      setPathing( &dc, PATH_GROUND_WITHROCKS,             false );
      setPathing( &dc, PATH_CWALK_NOROCKS,                false );
      setPathing( &dc, PATH_CWALK_WITHROCKS,              false );
      setPathing( &dc, PATH_BUILDABLE,                    false );
      setPathing( &dc, PATH_BUILDABLE_MAIN,               false );

    } else if( *type == "nobuildmain" ) {
      setPathing( &dc, PATH_BUILDABLE_MAIN,               false );

    } else {
      printError( "Placed object %s has unknown type %s.",
                  name->data(),
                  type->data() );
      exit( -1 );
    }
  }
}
